import logging
import random

from sqlalchemy import Column, Integer, String

from rollbot import RollbotFailure, as_plugin, ModelBase, pop_arg
from rollbot.plugins.injection import Database, Message, Subcommand, Bot, Config, Injector


class CurseBlessScore(ModelBase):
    __tablename__ = "curse_bless_score"
    group_id = Column(String, primary_key=True)
    person_id = Column(Integer, primary_key=True)
    curses = Column(Integer)
    blessings = Column(Integer)


def get_person_id(args):
    return args.strip().lower().replace(" ", "_")


def fmt_times(n):
    return str(n) + (" time" if n == 1 else " times")


def curse_agg(db: Database, msg: Message, args: Subcommand.ArgString):
    if args is None:
        return f"Sorry! I need to know who's records you're trying to aggregate!"
    person_id = get_person_id(args)
    curses = 0
    blessings = 0
    for cbs in db.query(CurseBlessScore).filter(CurseBlessScore.person_id == person_id):
        curses += cbs.curses
        blessings += cbs.blessings
    return f"From my records, {args} has been cursed {fmt_times(curses)} and blessed {fmt_times(blessings)} across all chats."


def curse_clear(db: Database, msg: Message, args: Subcommand.ArgString):
    if args is None:
        return f"Sorry! I need to know who's records you're trying to clear!"
    person_id = get_person_id(args)
    if not msg.from_admin:
        return RollbotFailure.PERMISSIONS
    cbs = db.query(CurseBlessScore).get((msg.group_id, person_id))
    if cbs is None:
        return f"Sorry! I don't know who {args} is!"
    oc = cbs.curses
    ob = cbs.blessings
    cbs.curses = 0
    cbs.blessings = 0
    return f"Done! I have cleared all records of {args} in this chat, who was previously cursed {fmt_times(oc)} and blessed {fmt_times(ob)}!"


def curse_clearall(db: Database, msg: Message, args: Subcommand.ArgString):
    if args is None:
        return f"Sorry! I need to know who's records you're trying to clear!"
    person_id = get_person_id(args)
    if not msg.from_admin:
        return RollbotFailure.PERMISSIONS
    cnt = 0
    oc = 0
    ob = 0
    for cbs in db.query(CurseBlessScore).filter(CurseBlessScore.person_id == person_id):
        oc += cbs.curses
        ob += cbs.blessings
        cbs.curses = 0
        cbs.blessings = 0
        cnt += 1
    if cnt == 0:
        return f"Sorry! I don't know who {args} is!"
    return f"Done! I have cleared all records of {args} in all chats, who was previously cursed {fmt_times(oc)} and blessed {fmt_times(ob)} in total!"


def curse_top(db: Database, msg: Message, args: Subcommand.ArgString):
    if msg.command == "curse":
        target = CurseBlessScore.curses
        text = "cursed"
        extract = lambda cbs: cbs.curses
    elif msg.command == "bless":
        target = CurseBlessScore.blessings
        text = "blessed"
        extract = lambda cbs: cbs.blessings
    else:
        return "Sorry! Right now !top can only be used with curse and bless."
    
    rankings = "\n".join(f"{i + 1}. {cbs.person_id.capitalize()} was {text} {fmt_times(extract(cbs))}"
        for i, cbs in enumerate(db.query(CurseBlessScore
            ).filter(CurseBlessScore.group_id == msg.group_id
            ).filter(target != 0
            ).order_by(target.desc()
            ).limit(10)))
    if len(rankings) == 0:
        return "Sorry! I don't have enough curse history in this chat to give statistics."
    return f"The worst offenders of this chat:\n{rankings}"


SUBC_MAP = {
    "aggregate": curse_agg,
    "clear": curse_clear,
    "clearall": curse_clearall,
    "top": curse_top,
}


def on_curse(score):
    # Note we do (x := x + 1) instead of += b/c that can create race conditions in sqlalchemy
    score.curses = score.curses + 1
    return score


def on_bless(score):
    score.blessings = score.blessings + 1
    return score


def on_uncurse(score):
    score.curses = score.curses - 1
    return score


def on_unbless(score):
    score.blessings = score.blessings - 1
    return score


CMD_MAP = {
    "curse": on_curse,
    "bless": on_bless,
    "blurse": lambda score: on_bless(on_curse(score)),
    "uncurse": on_uncurse,
    "unbless": on_unbless,
}


def curse_command(db: Database, 
                  msg: Message,
                  bot: Bot, 
                  ban_list: Config("curse.banlist"), 
                  no_list: Config("curse.nolist")):
    if not msg.from_admin and msg.command.startswith("un"):
        return RollbotFailure.PERMISSIONS.with_reason("For now, only admins can unbless or uncurse people!")

    if msg.raw_args is None:
        return f"Sorry - you need to provide the name of someone to {msg.command}!"

    person_id = get_person_id(msg.raw_args)
    if msg.sender_id in ban_list and person_id in ban_list[msg.sender_id]:
        if msg.sender_id in no_list:
            return "No!!!!!!!!!!"
        if random.randint(0, 1) == 0:
            bot.manually_post_message("Hey! You aren't allowed to affect that person's score! And cheaters never prosper!", msg.group_id)
            return f"!curse {ban_list[msg.sender_id][0]}"
        return "Hey! You aren't allowed to affect that person's score!"

    score = db.query(CurseBlessScore).get((msg.group_id, person_id))
    if score is None:
        score = CurseBlessScore(group_id=msg.group_id, person_id=person_id, curses=0, blessings=0)
        db.add(score)
    CMD_MAP[msg.command](score)

    return f"It is done! {msg.raw_args} has been cursed {fmt_times(score.curses)} and blessed {fmt_times(score.blessings)} in this chat."


@as_plugin
def curse(subc: Subcommand, injector: Injector):
    if subc is None:
        return injector.run_with_deps(curse_command)
    try:
        return injector.run_with_deps(SUBC_MAP[subc.command])
    except KeyError:
        return RollbotFailure.INVALID_SUBCOMMAND
