from random import randint, gauss, shuffle, choices
from datetime import datetime, timedelta
import pickle

from sudoku_py import SudokuGenerator, Sudoku

from rollbot import as_plugin, with_help, as_sender_singleton, as_group_singleton, RollbotFailure
from rollbot.plugins.decorators import with_startup, require_min_args, require_args, require_admin
from rollbot.plugins.injection import Config, Arg, Singleton, Message, Lazy, Query

@as_sender_singleton
class RollcoinWallet:
    balance: float = 0.0


@as_sender_singleton
class RollcoinBegCooldown:
    cooldown: datetime = datetime(2000, 1, 1)


@as_group_singleton
class RollcoinBlockchain:
    puzzle: "binary"

    def get_puzzle(self):
        if self.puzzle is None:
            return None
        return pickle.loads(self.puzzle)

    def set_puzzle(self, new_puzzle):
        self.puzzle = pickle.dumps(new_puzzle)
    
    def clear_puzzle(self):
        self.puzzle = None


@as_sender_singleton
class RollcoinGamblingWallet:
    balance: float = 0.0


@as_group_singleton
class RollcoinMarket:
    state: str = "neutral"


def setup_initial_balances(plugin, db):
    bot = plugin.bot
    bot.logger.info("Setting up rollcoin wallets")
    wallets = bot.config.get("rollcoin.wallets")
    initial = bot.config.get("rollcoin.initial")
    force = bot.config.get("rollcoin.force_initial")
    for k, v in initial.items():
        wallet_key = wallets.get(k, None)
        if wallet_key is None:
            continue
        wallet = db.query(RollcoinWallet).get(wallet_key)
        if wallet is None:
            wallet = RollcoinWallet.create_from_key(wallet_key)
        elif k not in force:
            continue
        wallet.balance = v
        db.add(wallet)
    bot.logger.info("Initial wallets created!")


@with_startup(setup_initial_balances)
@with_help("Check your balance of Rollcoins")
@as_plugin
def balance(wallet: Singleton(RollcoinWallet), holdings: Singleton(RollcoinGamblingWallet)):
    return f"You have {wallet.balance} Rollcoins in your wallet and {holdings.balance} Rollcoins in the market"


SPECIAL_AMOUNTS = ("ALL", "FRAC", "-ALL", "-FRAC", "HALF", "-HALF")
SPECIAL_MULTS = {
    "B": 10 ** 9,
    "M": 10 ** 6,
    "K": 10 ** 3,
    "T": 10 ** 3,
}


def parse_amount(amount):
    if (up_amount := amount.upper()) in SPECIAL_AMOUNTS:
        return up_amount
    mult = SPECIAL_MULTS.get(up_amount[-1], None)
    if mult is None:
        return float(amount)
    else:
        return mult * float(amount[:-1])


def assert_positive(amount):
    if amount <= 0:
        RollbotFailure.INVALID_ARGUMENTS.with_reason(f"Amount should be positive").raise_exc()


def fractional_part(number):
    return float("0." + str(float(number)).split(".")[1])


def lookup_target(target_name: Arg(0), wallets: Config("rollcoin.wallets")):
    target_id = wallets.get(target_name.lower(), None)
    if target_id is None:
        RollbotFailure.INVALID_ARGUMENTS.with_reason(f"Could not find wallet-holder {target}").raise_exc()
    
    return target_id


@with_help("Tip someone else some Rollcoins: !tip target amount")
@require_min_args(2, "Tip requires 2 arguments - target and amount")
@as_plugin
def tip(target_name: Arg(0),
        amount: Arg(1, parse_amount, "Amount should be a number or ALL or FRAC - not {}"),
        sender_wallet: Singleton(RollcoinWallet), 
        target_wallet: Lazy(Singleton(RollcoinWallet).by(lookup_target))):
    if amount in SPECIAL_AMOUNTS:
        if amount == "ALL":
            amount = sender_wallet.balance
        elif amount == "FRAC":
            amount = fractional_part(sender_wallet.balance)
        elif amount == "HALF":
            amount = sender_wallet.balance / 2
        else:
            amount = -1
    assert_positive(amount)

    if sender_wallet.balance < amount:
        return f"Insufficient funds! You need {round(amount - sender_wallet.balance, 2)} more Rollcoins. Have you tried to !mine, !beg, or !gamble?"

    target_wallet = target_wallet.get()
    if target_wallet.sender_id == sender_wallet.sender_id:
        return f"You can't tip yourself!"

    sender_wallet.balance = sender_wallet.balance - amount
    target_wallet.balance = target_wallet.balance + amount
    return f"Done! {target_name} now has {target_wallet.balance} Rollcoins"


def get_non_sender_ids(wallets: Config("rollcoin.wallets"), msg: Message):
    return [w for w in wallets.values() if w != msg.sender_id]


@with_help("Donate money to be distributed evenly among all wallets")
@require_min_args(1, "Need an amount to donate")
@as_plugin
def donate(amount: Arg(0, parse_amount, "Amount should be a number or ALL or FRAC - not {}"), 
           sender_wallet: Singleton(RollcoinWallet),
           other_wallets: Lazy(Singleton(RollcoinWallet).by_all(get_non_sender_ids))):
    if amount in SPECIAL_AMOUNTS:
        if amount == "ALL":
            amount = sender_wallet.balance
        elif amount == "FRAC":
            amount = fractional_part(sender_wallet.balance)
        elif amount == "HALF":
            amount = sender_wallet.balance / 2
        else:
            amount = -1
    assert_positive(amount)

    if sender_wallet.balance < amount:
        return f"Insufficient funds! You need {round(amount - sender_wallet.balance, 2)} more Rollcoins"

    other_wallets = other_wallets.get()
    per_person = amount / (len(other_wallets))
    for target_wallet in other_wallets:
        target_wallet.balance = target_wallet.balance + per_person
    sender_wallet.balance = sender_wallet.balance - amount

    return f"Done! You have donated {per_person} to each of the {len(other_wallets)} other wallet(s)"


@with_help("Receive a mining question you can answer to receive Rollcoins")
@as_plugin
def mine(msg: Message, 
         blockchain: Singleton(RollcoinBlockchain), 
         wallet: Singleton(RollcoinWallet),
         market_rates: Config("rollcoin.mining"),
         market_state: Lazy(Singleton(RollcoinMarket))):
    skip = msg.raw_args is not None and msg.raw_args.strip().lower() == "skip" and msg.from_admin

    if skip:
        blockchain.clear_puzzle()
    
    puzzle = blockchain.get_puzzle()
    if puzzle is None:
        exchange = list(zip("abcdefghi", range(10)))
        shuffle(exchange)
        gen = SudokuGenerator(9)
        gen.generate(0)
        gen.board_exchange_values({k: v + 1 for k, v in exchange})
        gen.generate(randint(10, 50))
        puzzle = gen.board
        blockchain.set_puzzle(puzzle)

    if skip:
        return "Admin has skipped the previous mining challenge.\nThe current mining challenge is\n" + str(Sudoku(board=puzzle))

    if msg.raw_args is None:
        return "The current mining challenge is\n" + str(Sudoku(board=puzzle))

    parsed = []
    row = []
    for c in msg.raw_args:
        if not c.isdigit():
            continue
        row.append(int(c))
        if len(row) == 9:
            parsed.append(row)
            row = []
    
    try:
        guess = Sudoku(board=parsed, block_width=3, block_height=3)
    except:
        return "Sorry, I could not parse that solution!"

    for i, row in enumerate(puzzle):
        for j, cell in enumerate(row):
            if cell not in (0, parsed[i][j]):
                return "Sorry, that solution doesn't match the puzzle! The current mining challenge is\n" + str(Sudoku(board=puzzle))
    
    for row in parsed:
        if set(row) != set(range(1, 10)):
            return "Sorry, that solution isn't valid!"
    
    for i in range(9):
        if set(r[i] for r in parsed) != set(range(1, 10)):
            return "Sorry, that solution isn't valid!"

    for i in range(0, 9, 3):
        for j in range(0, 9, 3):
            if set(x for r in parsed[i:i+3] for x in r[j:j+3]) != set(range(1, 10)):
                return "Sorry, that solution isn't valid!"

    zeroes = len([y for x in puzzle for y in x if y == 0])
    difficulty = (zeroes - 10) / 4
    market_rate = market_rates.get(market_state.get().state, 10)
    actual_value = round(abs(gauss(market_rate + difficulty, 1)), 2)
    wallet.balance = wallet.balance + actual_value
    blockchain.clear_puzzle()
    return f"Looks right to me! The current market rate is {actual_value} Rollcoins per Sudoku, so that brings your balance to {wallet.balance}"


@with_help("Beg for some Rollcoins")
@as_plugin
def beg(wallet: Singleton(RollcoinWallet), 
        beg_cooldown: Singleton(RollcoinBegCooldown), 
        cooldown_time: Config("rollcoin.cooldown"), 
        beg_chance: Config("rollcoin.beg_chance")):
    old_cooldown = beg_cooldown.cooldown
    beg_cooldown.cooldown = datetime.now() + timedelta(**cooldown_time)

    if datetime.now() < old_cooldown:
        return f"Don't beg so frequently - it's not a good look"

    if randint(0, 99) > beg_chance:
        return "What are you, poor?"

    increase = round(abs(gauss(1, 1)), 2)
    wallet.balance = wallet.balance + increase
    return f"Fine, I've given you {increase} Rollcoins, which brings your balance to {wallet.balance}"


def not_sender_gambling_wallet(msg: Message):
    return RollcoinGamblingWallet.sender_id != msg.sender_id


@with_help("Gamble some Rollcoins")
@require_args(1, "Gambling requires exactly one argument: the amount to transfer (can be positive or negative)")
@as_plugin
def gamble(amount: Arg(0, parse_amount, "Amount should be a number or ALL or FRAC - not {}"), 
           sender_wallet: Singleton(RollcoinWallet), 
           sender_holdings: Singleton(RollcoinGamblingWallet), 
           market: Singleton(RollcoinMarket), 
           transitions: Config("rollcoin.market"),
           other_holdings: Lazy(Query(RollcoinGamblingWallet).filter(not_sender_gambling_wallet))):
    # pos moves money wallet -> gambling, neg moves money gambling -> wallet
    # market has a current state (neutral, bull, bear, boom, bust, collapse, etc.)
    # after a transfer is made, the market evolves based on
    #   - direction of transfer (buy vs sell)
    #   - current market state
    #   - rng
    # market enters new state and affects all gambling wallets, multiplying their balances
    if amount in SPECIAL_AMOUNTS:
        if amount == "ALL":
            amount = sender_wallet.balance
        elif amount == "FRAC":
            amount = fractional_part(sender_wallet.balance)
        elif amount == "HALF":
            amount = sender_wallet.balance / 2
        elif amount == "-ALL":
            amount = -sender_holdings.balance
        elif amount == "-FRAC":
            amount = -fractional_part(sender_holdings.balance)
        elif amount == "-HALF":
            amount = -sender_holdings.balance / 2
        else:
            raise ValueError(amount)

    if amount == 0:
        RollbotFailure.INVALID_ARGUMENTS.with_reason(f"Amount should be nonzero").raise_exc()

    buying = amount > 0
    abs_amount = abs(amount)
    if buying and abs_amount > sender_wallet.balance:
        return f"Insufficient funds! You need {round(abs_amount - sender_wallet.balance, 2)} more Rollcoins. Have you tried to !mine, !beg, or !gamble?"
    elif not buying and abs_amount > sender_holdings.balance:
        return f"Insufficient holdings! You would need {round(abs_amount - sender_holdings.balance, 2)} more Rollcoins in the market."

    options = transitions[market.state]["buy" if buying else "sell"]
    transition = choices(options, weights=tuple(x["weight"] for x in options))[0]
    multiplier = transition["multiplier"]
    market.state = transition["new_state"]
    sender_wallet.balance = sender_wallet.balance - amount
    sender_holdings.balance = multiplier * (sender_holdings.balance + amount)

    for holdings in other_holdings.get():
        holdings.balance = multiplier * holdings.balance

    return f"{transition['message']}\nYour current holdings: {sender_holdings.balance}"


@with_help("View the whole blockchain (i.e., who has how many rollcoins)")
@as_plugin
def blockchain(
    names: Config("rollcoin.wallets"),
    wallets: Query(RollcoinWallet),
    holdings: Query(RollcoinGamblingWallet)):
    wallets = {w.sender_id: w.balance for w in wallets}
    holdings = {w.sender_id: w.balance for w in holdings}
    results = []
    for name, wallet_id in names.items():
        wallet = wallets.get(wallet_id, None)
        market = holdings.get(wallet_id, None)
        status = name.title() + "\n"
        total = 0
        if wallet is not None:
            status += f"\t\tWallet: {wallet}\n"
            total += wallet
        else:
            status += "\t\tHas not used their rollcoin wallet\n"
        if market is not None:
            status += f"\t\tMarket: {market}\n"
            total += market
        else:
            status += "\t\tHas not invested their rollcoins\n"
        if total != 0:
            status += f"\t\tTotal: {total}\n"
        results.append((total, name, status))
    return "\n".join(r for _, _, r in sorted(results, reverse=True))


@with_help("Deflate the entire economy")
@require_args(1, "Deflating requires exactly one argument: power of 10 to deflate by")
@require_admin
@as_plugin
def deflate(factor: Arg(0, int, "Deflation factor must be an integer - not {}"), 
            wallets: Query(RollcoinWallet), 
            holdings: Query(RollcoinGamblingWallet)):
    assert_positive(factor)
    actual_factor = 10 ** -factor
    for w in wallets:
        w.balance = w.balance * actual_factor
    for w in holdings:
        w.balance = w.balance * actual_factor
    return f"Done! I have deflated the entire economy by a factor of 10^{factor}"
